from django.core.exceptions import ObjectDoesNotExist

from utility.exceptions import ObjectNotFoundError
from utility.functions import contains_char_classes, is_email_valid


class GenericRepository():
    """
    Provides a generic database abstraction layer that does not depend on
    the model class
    """
    def __init__(self, model_class):
        """
        Constructor
        """
        self.model_class = model_class

    def find_by_user(self, user):
        """
        Provides generic find by user function
        """
        if hasattr(self.model_class, 'owner'):
            return self.model_class.objects.filter(owner=user)

        try:
            return self.model_class.objects.filter(id=user.id)
        except AttributeError:
            raise ObjectNotFoundError('No object matching user found.',
                                      modelclass=self.model_class)

    def find_by_id(self, pk, user):
        """
        Provides generic find by id function
        """
        if hasattr(self.model_class, 'owner'):
            return self.model_class.objects.filter(owner=user).filter(id=pk)

        try:
            return self.model_class.objects.filter(id=user.id).filter(id=pk)
        except AttributeError:
            raise ObjectNotFoundError('No object matching user found.',
                                      modelclass=self.model_class)

    def get_by_id(self, pk, user):
        """
        Provides generic get by id function
        """
        if hasattr(self.model_class, 'owner'):
            try:
                return self.model_class.objects.filter(owner=user).get(id=pk)
            except ObjectDoesNotExist:
                raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

        try:
            return self.model_class.objects.filter(id=user.id).get(id=pk)
        except AttributeError:
            raise ObjectNotFoundError('No object matching user found.',
                                      modelclass=self.model_class)
        except ObjectDoesNotExist:
            raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

    def get_by_user_and_id(self, pk, user):
        """
        Provides generic get by user and id function
        """
        if hasattr(self.model_class, 'owner'):
            try:
                return self.model_class.objects.filter(owner=user).get(id=pk)
            except ObjectDoesNotExist:
                raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

        try:
            return self.model_class.objects.filter(id=user.id).get(id=pk)
        except AttributeError:
            raise ObjectNotFoundError('No object matching user found.',
                                      modelclass=self.model_class)
        except ObjectDoesNotExist:
            raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

    def list_all(self):
        """
        Provides generic list all function
        """
        return self.model_class.objects.all()

    def delete_by_id(self, pk, user):
        """
        Provides generic delete function
        """
        if hasattr(self.model_class, 'owner'):
            try:
                return self.model_class.objects.filter(owner=user).get(id=pk).delete()
            except ObjectDoesNotExist:
                raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

        try:
            return self.model_class.objects.filter(id=user.id).get(id=pk).delete()
        except AttributeError:
            raise ObjectNotFoundError('No object matching user found.',
                                      modelclass=self.model_class)
        except ObjectDoesNotExist:
            raise ObjectNotFoundError('Object not found.', modelclass=self.model_class)

    @staticmethod
    def persist(instance):
        """
        Provides generic persist function
        """
        return instance.save()

    def check_unique(self, pk, field, value):
        """
        Checks the uniqueness of value
        """
        if value is None:
            return 0

        kwargs = {field: str(value)}
        return self.model_class.objects.exclude(id=pk).filter(**kwargs)

    @staticmethod
    def check_password_too_short(password, length):
        """
        Checks if password is at least as long as length demands
        """
        if len(password) < length:
            return True

        return False

    @staticmethod
    def check_too_few_char_classes(password, noc):
        """
        Checks if password contains the demanded number of different
        char classes
        """
        if not contains_char_classes(password, noc):
            return True

        return False

    @staticmethod
    def check_email_valid(email):
        """
        Checks if an email address is valid
        """
        return email is None or is_email_valid(email)



from django.db.models.query import QuerySet
from utility.exceptions import ValidationError


def is_value_unique(repository, field):
    """
    Checks if a value is unique
    """
    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            i_dir = vars(instance)
            if repository.check_unique(instance.id, field, i_dir[field]):
                raise ValidationError('This value has alredy been used.', field=field)
            return True

    return _checker


def validate_many(*validators):
    """
    Allows to apply multiple validators
    """
    def _checker(instance):
        return all(validator(instance) for validator in validators)

    return _checker



from utility.functions import get_uuid1


def generic_factory(model_class, **kwargs):
    """
    Creates and returns a new instance with an automatically generated id
    """
    return model_class(id=get_uuid1(), **kwargs)



from django.contrib.auth.hashers import make_password


def generic_retrieve_single_service(repository):
    """
    Returns an instance by user and pk
    """
    def _retrieve(pk, user):
        instance = repository.get_by_id(pk, user)
        return instance

    return _retrieve


def generic_retrieve_all_service(repository):
    """
    Returns all stored instances
    """
    def _retrieve(user=None):
        instances = repository.list_all()
        return instances

    return _retrieve


def generic_retrieve_all_by_owner_service(repository):
    """
    Returns all stored instances owned by user
    """
    def _retrieve(user):
        instances = repository.find_by_user(user)
        return instances

    return _retrieve


def generic_delete_service(repository):
    """
    Deletes an instance by pk
    """
    def _delete(pk, user):
        instance = repository.get_by_id(pk, user)
        repository.delete_by_id(pk, user)
        return instance

    return _delete


def generic_update_service(repository, validator):
    """
    Updates an existing instance
    """
    def _update(pk, user, **kwargs):
        instance = repository.get_by_user_and_id(pk, user)
        for key, value in kwargs.items():
            instance[key] = value
        validator(instance)

        if hasattr(repository.model_class, 'password') and 'password' in kwargs:
            instance.password = make_password(kwargs['password'])

        repository.persist(instance)
        return instance

    return _update


def generic_create_service(repository, validator, factory):
    """
    Returns and persists a new instance
    """
    def _create(user, **kwargs):
        if hasattr(repository.model_class, 'owner'):
            kwargs['owner'] = user

        instance = factory(repository.model_class, **kwargs)
        validator(instance)

        if hasattr(repository.model_class, 'password') and 'password' in kwargs:
            instance.password = make_password(kwargs['password'])

        repository.persist(instance)
        return instance

    return _create



from rest_framework import viewsets

from tests.serializers import UserObjectSerializer
from tests.models import UserObject

from generic.views import GenericViewset
from generic.factories import generic_factory
from generic.repositories import GenericRepository
from generic.validators import validate_many, is_value_unique
from generic.services import (generic_retrieve_single_service,
                              generic_retrieve_all_service,
                              generic_delete_service,
                              generic_update_service,
                              generic_create_service)


class ObjectViewSet(viewsets.ModelViewSet):
    repository = GenericRepository(UserObject)
    serializer_class = UserObjectSerializer

    def retrieve(self, request, pk=None, format=None):
        service = generic_retrieve_single_service(self.repository)

        view = GenericViewset(self.serializer_class, service, request)
        return view.retrieve(pk)

    def list(self, request, format=None):
        service = generic_retrieve_all_service(self.repository)

        view = GenericViewset(self.serializer_class, service, request)
        return view.list()

    def destroy(self, request, pk=None, format=None):
        service = generic_delete_service(self.repository)

        view = GenericViewset(self.serializer_class, service, request)
        return view.destroy(pk)

    def partial_update(self, request, pk=None, format=None):
        validators = (is_value_unique(self.repository, 'name'),)
        service = generic_update_service(self.repository, validate_many(*validators))

        view = GenericViewset(self.serializer_class, service, request)
        return view.partial_update(pk)

    def create(self, request, format=None):
        validators = (is_value_unique(self.repository, 'name'),)
        service = generic_create_service(self.repository, validate_many(*validators),
                                         generic_factory)

        view = GenericViewset(self.serializer_class, service, request)
        return view.create()



from django.db import models
from django.contrib.auth.models import AbstractUser

from utility.functions import get_uuid1
from utility.mixins import DictMixin


class User(AbstractUser, DictMixin):
    """
    Overrides the default Django User, uuid is changed from an auto incremental
    integer field to a uuidv1 field
    """
    id = models.UUIDField(default=get_uuid1, primary_key=True)
    username = models.CharField(max_length=50, unique=True, blank=False, null=False)
    password = models.CharField(max_length=255, blank=False, null=False)
    email = models.EmailField(blank=True, null=True)
    first_name = models.CharField(max_length=1, blank=True, null=True)
    last_name = models.CharField(max_length=1, blank=True, null=True)
    firstName = models.CharField(max_length=50, blank=True, null=True)
    lastName = models.CharField(max_length=50, blank=True, null=True)
    validated_email = models.BooleanField(default=False, blank=False, null=False)
    isFacilityManagement = models.BooleanField(default=False, blank=False, null=False)
    isProprietor = models.BooleanField(default=False, blank=False, null=False)
    isTenant = models.BooleanField(default=False, blank=False, null=False)

    def __unicode__(self):
        return "{0}".format(self.id)

    def __str__(self):
        return "{0}".format(self.id)


class Object(models.Model, DictMixin):
    id = models.UUIDField(default=get_uuid1, primary_key=True)
    name = models.CharField(unique=True, max_length=255, blank=False, null=False)
    password = models.CharField(unique=False, max_length=255, blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'object'


class UserObject(models.Model, DictMixin):
    id = models.UUIDField(default=get_uuid1, primary_key=True)
    owner = models.ForeignKey('User', models.CASCADE, related_name='user_objects')
    name = models.CharField(unique=True, max_length=255, blank=False, null=False)

    class Meta:
        managed = True
        db_table = 'user_object'


class UserObjectPw(models.Model, DictMixin):
    id = models.UUIDField(default=get_uuid1, primary_key=True)
    owner = models.ForeignKey('User', models.CASCADE, related_name='user_objects_pw')
    name = models.CharField(unique=True, max_length=255, blank=False, null=False)
    password = models.CharField(unique=False, max_length=255, blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'user_object_pw'
